"""The constants for graphical plot style."""

COLORS = [
    "#ef597b",  # bright pink
    "#29a2c6",  # cyan blue
    "#ffcb18",  # vivid yellow
    "#73b66b",  # medium green
    "#4e4f4b",  # dark gray
    "#a23e9a",  # purple
    "#f59331",  # orange
    "#007f5f",  # deep teal green
    "#00509e",  # deep blue
    "#c1b7b0",  # light beige gray
    "#ff6f00",  # strong orange
    "#0089a3",  # ocean blue
    "#b2df8a",  # light lime green
    "#cab2d6",  # soft lavender
    "#6a3d9a",  # dark violet
    "#fb9a99",  # soft pink
    "#1b9e77",  # forest green
    "#d95f02",  # burnt orange
    "#7570b3",  # indigo
    "#e7298a",  # magenta
]
FIGSIZE = (10.0, 8.0)  # (width, height)
MARKERS = [
    "o",
    "x",
    "+",
    "^",
    "d",
    "s",
    "v",
    "*",
    "D",
    "H",
    "p",
    "<",
    ">",
    "1",
    "2",
]  # list of dot shapes
MARKERSIZE = 4
